"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetadataGenerator = void 0;
var glob_1 = require("glob");
var lodash_1 = require("lodash");
var typescript_1 = require("typescript");
var debug_1 = require("../debug");
var decoratorUtils_1 = require("../utils/decoratorUtils");
var controllerGenerator_1 = require("./controllerGenerator");
var resolver_1 = require("./resolver");
var M = require("minimatch");
var MetadataGenerator = /** @class */ (function () {
    function MetadataGenerator(entryFile, compilerOptions, ignorePaths) {
        this.ignorePaths = ignorePaths;
        this.nodes = new Array();
        this.referenceTypes = {};
        this.circularDependencyResolvers = new Array();
        this.debugger = debug_1.useDebugger();
        resolver_1.TypeNodeResolver.clearCache();
        var sourceFiles = this.getSourceFiles(entryFile);
        this.debugger('Starting Metadata Generator');
        this.debugger('Source files: %j ', sourceFiles);
        this.debugger('Compiler Options: %j ', compilerOptions);
        this.program = typescript_1.createProgram(sourceFiles, compilerOptions);
        this.typeChecker = this.program.getTypeChecker();
    }
    MetadataGenerator.prototype.generate = function () {
        var _this = this;
        this.program.getSourceFiles().forEach(function (sf) {
            if (_this.ignorePaths && _this.ignorePaths.length) {
                for (var _i = 0, _a = _this.ignorePaths; _i < _a.length; _i++) {
                    var path = _a[_i];
                    if (
                    // sf.fileName.includes('node_modules/typescript-rest/') ||
                    // sf.fileName.includes('node_modules/@decorators/express') ||
                    M(sf.fileName, path)) {
                        return;
                    }
                }
            }
            typescript_1.forEachChild(sf, function (node) {
                if (typescript_1.isModuleDeclaration(node)) {
                    /**
                     * For some reason unknown to me, TS resolves both `declare module` and `namespace` to
                     * the same kind (`ModuleDeclaration`). In order to figure out whether it's one or the other,
                     * we check the node flags. They tell us whether it is a namespace or not.
                     */
                    // tslint:disable-next-line:no-bitwise
                    if ((node.flags & typescript_1.NodeFlags.Namespace) === 0 && node.body && typescript_1.isModuleBlock(node.body)) {
                        node.body.statements.forEach(function (statement) {
                            _this.nodes.push(statement);
                        });
                        return;
                    }
                }
                _this.nodes.push(node);
            });
        });
        this.debugger('Building Metadata for controllers Generator');
        var controllers = this.buildControllers();
        this.debugger('Handling circular references');
        this.circularDependencyResolvers.forEach(function (c) { return c(_this.referenceTypes); });
        return {
            controllers: controllers,
            referenceTypes: this.referenceTypes
        };
    };
    MetadataGenerator.prototype.isExportedNode = function (node) {
        return true;
    };
    MetadataGenerator.prototype.addReferenceType = function (referenceType) {
        if (!referenceType.refName) {
            return;
        }
        this.referenceTypes[referenceType.refName] = referenceType;
    };
    MetadataGenerator.prototype.getReferenceType = function (refName) {
        return this.referenceTypes[refName];
    };
    MetadataGenerator.prototype.onFinish = function (callback) {
        this.circularDependencyResolvers.push(callback);
    };
    MetadataGenerator.prototype.getClassDeclaration = function (className) {
        var found = this.nodes
            .filter(function (node) {
            var classDeclaration = node;
            return (node.kind === typescript_1.SyntaxKind.ClassDeclaration && classDeclaration.name && classDeclaration.name.text === className);
        });
        if (found && found.length) {
            return found[0];
        }
        return undefined;
    };
    MetadataGenerator.prototype.getInterfaceDeclaration = function (className) {
        var found = this.nodes
            .filter(function (node) {
            var interfaceDeclaration = node;
            return (node.kind === typescript_1.SyntaxKind.InterfaceDeclaration && interfaceDeclaration.name && interfaceDeclaration.name.text === className);
        });
        if (found && found.length) {
            return found[0];
        }
        return undefined;
    };
    MetadataGenerator.prototype.getSourceFiles = function (sourceFiles) {
        var _this = this;
        this.debugger('Getting source files from expressions');
        this.debugger('Source file patterns: %j ', sourceFiles);
        var sourceFilesExpressions = lodash_1.castArray(sourceFiles);
        var result = new Set();
        var options = { cwd: process.cwd() };
        sourceFilesExpressions.forEach(function (pattern) {
            _this.debugger('Searching pattern: %s with options: %j', pattern, options);
            var matches = glob_1.sync(pattern, options);
            matches.forEach(function (file) { result.add(file); });
        });
        return Array.from(result);
    };
    MetadataGenerator.prototype.buildControllers = function () {
        var _this = this;
        return this.nodes
            .filter(function (node) { return node.kind === typescript_1.SyntaxKind.ClassDeclaration; })
            .filter(function (node) { return !decoratorUtils_1.isDecorator(node, function (decorator) { return 'Hidden' === decorator.text; }); })
            .filter(function (node) { return decoratorUtils_1.isDecorator(node, function (decorator) { return decorator.text === 'Path' || decorator.text === 'Controller'; }); })
            .map(function (classDeclaration) { return new controllerGenerator_1.ControllerGenerator(classDeclaration, _this); })
            .filter(function (generator) { return generator.isValid(); })
            .map(function (generator) { return generator.generate(); });
    };
    return MetadataGenerator;
}());
exports.MetadataGenerator = MetadataGenerator;
//# sourceMappingURL=metadataGenerator.js.map